/*
Module : BITSWrappers.h
Purpose: Defines the interface for a set of C++ class which encapsulate the Windows BITS APIs.
History: PJN / 06-08-2013 1. Initial creation
         PJN / 06-08-2013 1. Provided a CJob::GetOwner wrapper method
                          2. Provided a CJob::GetProxySettings wrapper method
         PJN / 10-08-2013 1. Added support for the IBitsTokenOptions interface
         PJN / 15-01-2015 1. Updated copyright details.
                          2. Added SAL annotations to all the code
                          3. Verified the code compiles cleanly in VC 2015
         PJN / 29-04-2017 1. Updated copyright details.
                          2. Updated the code to compile cleanly using /permissive-.
         PJN / 29-10-2017 1. Implemented support for IBackgroundCopyCallback2 via new CCallback2 class
                          2. Implemented support for IBackgroundCopyCallback3 via new CCallback3 class
                          3. Implemented support for IBackgroundCopyFile5 via new CFile5 class.
                          4. Implemented support for IBackgroundCopyFile6 via new CFile6 class.
                          5. Fixed an issue in CJobHttpOptions::GetClientCertificate where the wrapper code
                          did not correctly handle a S_FALSE return value from the low level API.
                          6. Fixed an issue in CJob::SetNotifyInterface related to getting IUnknown interface
                          prior to passing to calling SetNotifyInterface low level API.
                          7. Replaced NULL with nullptr throughout the codebase. This means that the code now 
                          requires VC 2010 at a minimum to compile.
                          8. Replaced BOOL throughout the codebase with bool.
                          9. Replaced CString::operator LPC*STR() calls with CString::GetString calls
         PJN / 07-07-2018 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean that 
                          the code will now only compile on VC 2017 or later.
                          3. Remove the code path which supported CBITSWRAPPERS_MFC_EXTENSIONS
         PJN / 16-09-2018 1. Fixed a number of compiler warnings when using VS 2017 15.8.4
         PJN / 18-11-2018 1. Implemented support for IBackgroundCopyJobHttpOptions2 via new CJobOptions2 class
                          2. Implemented support for IBackgroundCopyServerCertificateValidationCallback via 
                          new CServerCertificateValidationCallback class.
                          3. Renamed CJobHttpOptions class to CJobOptions
                          4. Implemented support for IBackgroundCopyJobHttpOptions3 via new CJobOptions3 class
         PJN / 19-04-2019 1. Updated copyright details
                          2. Updated the code to clean compile on VC 2019
         PJN / 15-09-2019 1. Fixed a number of compiler warnings when the code is compiled with VS 2019 Preview
         PJN / 10-03-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 12-04-2020 1. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 16-03-2022 1. Updated copyright details
                          2. Updated the code to use C++ uniform initialization for all variable declarations.

Copyright (c) 2013 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


/////////////////////////// Macros / Includes /////////////////////////////////

#pragma once

#ifndef __BITSWRAPPERS_H__
#define __BITSWRAPPERS_H__

#ifndef CBITSWRAPPERS_EXT_CLASS
#define CBITSWRAPPERS_EXT_CLASS
#endif //#ifndef CBITSWRAPPERS_EXT_CLASS

#ifndef CBITSWRAPPERS_EXT_API
#define CBITSWRAPPERS_EXT_API
#endif //#ifndef CBITSWRAPPERS_EXT_API

#ifndef __ATLCOM_H__
#include <atlcom.h>
#pragma message("To avoid this message, please put atlcom.h in your pre compiled header (normally stdafx.h)")
#endif //#ifndef __ATLCOM_H__

#ifndef __bits_h__
#include <Bits.h>
#pragma message("To avoid this message, please put Bits.h in your pre compiled header (normally stdafx.h)")
#endif //#ifndef __bits_h__

#ifndef _VECTOR_
#include <vector>
#pragma message("To avoid this message, please put vector in your pre compiled header (normally stdafx.h)")
#endif //#ifndef _VECTOR_

#ifndef _STRING_
#include <string>
#pragma message("To avoid this message, please put string in your pre compiled header (normally stdafx.h)")
#endif //#ifndef _STRING_

#pragma comment(lib, "Bits.lib")


/////////////////////////// Classes ///////////////////////////////////////////

namespace BITS
{
//Forward declarations
  class CJob;
  class CFile;

//Typedefs
  using String = std::wstring;
  using FilesArray = std::vector<CFile>;
  using JobsArray = std::vector<CJob>;
  using FileRangeArray = std::vector<BG_FILE_RANGE>;
  using ByteArray = std::vector<BYTE>;

//Functions
inline void _HandleStringParameter(_In_ LPWSTR pszValue, _Inout_ String& sValue)
{
  //Update the output parameter
  sValue.clear();

  if (pszValue != nullptr)
  {
    sValue = pszValue;

    //Free the memory before we return
    CoTaskMemFree(pszValue);
  }
}

inline void _HandleStringParameter2(_In_ LPWSTR pszValue, _Inout_ String& sValue)
{
  if (pszValue != nullptr)
  {
    sValue = pszValue;

    //Free the memory before we return
    CoTaskMemFree(pszValue);
  }
}

//Wrapper for IBackgroundCopyFile3
class CFile3 : public ATL::CComPtr<IBackgroundCopyFile3>
{
public:
//Methods
  HRESULT GetTemporaryName(_Out_ String& sFileName)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

    //Call the low level API
    LPWSTR pwszFileName{nullptr};
    const HRESULT hr{p->GetTemporaryName(&pwszFileName)};
    if (FAILED(hr))
      return hr;

    //Handle updating the output parameter
    _HandleStringParameter(pwszFileName, sFileName);

    return hr;
  }
};

//Wrapper for IBackgroundCopyFile4
__if_exists(IBackgroundCopyFile4)
{
class CFile4 : public ATL::CComPtr<IBackgroundCopyFile4>
{
public:
//No Helper methods defined
};
} //__if_exists(IBackgroundCopyFile4)

//Wrapper for IBackgroundCopyFile5
__if_exists(IBackgroundCopyFile5)
{
class CFile5 : public ATL::CComPtr<IBackgroundCopyFile5>
{
public:
//No Helper methods defined
};
} //__if_exists(IBackgroundCopyFile5)

//Wrapper for IBackgroundCopyFile6
__if_exists(IBackgroundCopyFile6)
{
class CFile6 : public ATL::CComPtr<IBackgroundCopyFile6>
{
public:
  HRESULT GetFilledFileRanges(_Out_ FileRangeArray& ranges)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

    //Call the low level API
    DWORD dwRangeCount{0};
    BG_FILE_RANGE* pRanges{nullptr};
    const HRESULT hr{p->GetFilledFileRanges(&dwRangeCount, &pRanges)};
    if (FAILED(hr))
      return hr;

    //Allocate the vector
    ranges.resize(dwRangeCount);
    if (pRanges != nullptr) //Handle the case where GetFilledFileRanges return S_FALSE with a null value for pRanges
    {
      int i{0};
      for (auto& range : ranges)
      {
#pragma warning(suppress: 6385 26481)
        range.InitialOffset = pRanges[i].InitialOffset;
#pragma warning(suppress: 26481)
        range.Length = pRanges[i].Length;
        ++i;
      }

      //Free the memory before we return
      CoTaskMemFree(pRanges);
    }

    return hr;
  }
};
} //__if_exists(IBackgroundCopyFile6)

//Wrapper for IBackgroundCopyFile2
class CFile2 : public ATL::CComPtr<IBackgroundCopyFile2>
{
public:
//Methods
  HRESULT GetFileRanges(_Out_ FileRangeArray& ranges)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

    //Call the low level API
    DWORD dwRangeCount{0};
    BG_FILE_RANGE* pRanges{nullptr};
    const HRESULT hr{p->GetFileRanges(&dwRangeCount, &pRanges)};
    if (FAILED(hr))
      return hr;

    //Allocate the vector
    ranges.resize(dwRangeCount);
    if (pRanges != nullptr) //Handle the case where GetFileRanges return S_FALSE with a null value for pRanges
    {
      int i{0};
      for (auto& range : ranges)
      {
#pragma warning(suppress: 6385 26481)
        range.InitialOffset = pRanges[i].InitialOffset;
#pragma warning(suppress: 26481)
        range.Length = pRanges[i].Length;
        ++i;
      }

      //Free the memory before we return
      CoTaskMemFree(pRanges);
    }

    return hr;
  }
};

//Wrapper for IBackgroundCopyFile
class CFile : public ATL::CComPtr<IBackgroundCopyFile>
{
public:
//Methods
  HRESULT GetLocalName(_Out_ String& sName)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

    //Call the low level API
    LPWSTR pwszName{nullptr};
    const HRESULT hr{p->GetLocalName(&pwszName)};
    if (FAILED(hr))
      return hr;

    //Handle updating the output parameter
    _HandleStringParameter(pwszName, sName);

    return hr;
  }

  HRESULT GetRemoteName(_Out_ String& sName)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

    //Call the low level API
    LPWSTR pwszName{nullptr};
    const HRESULT hr{p->GetRemoteName(&pwszName)};
    if (FAILED(hr))
      return hr;

    //Handle updating the output parameter
    _HandleStringParameter(pwszName, sName);

    return hr;
  }

  HRESULT GetFile2(_Inout_ CFile2& file2) noexcept
  {
    return QueryInterface(&file2);
  }

  HRESULT GetFile3(_Inout_ CFile3& file3) noexcept
  {
    return QueryInterface(&file3);
  }

__if_exists(CFile4)
{
  HRESULT GetFile4(_Inout_ CFile4& file4) noexcept
  {
    return QueryInterface(&file4);
  }
} //__if_exists(CFile4)

__if_exists(CFile5)
{
  HRESULT GetFile5(_Inout_ CFile5& file5) noexcept
  {
    return QueryInterface(&file5);
  }
} //__if_exists(CFile5)

__if_exists(CFile6)
{
  HRESULT GetFile6(_Inout_ CFile6& file6) noexcept
  {
    return QueryInterface(&file6);
  }
} //__if_exists(CFile6)
};

//Wrapper for IBackgroundCopyJob2
class CJob2 : public ATL::CComPtr<IBackgroundCopyJob2>
{
public:
//Methods
  HRESULT GetNotifyCmdLine(_Out_ String& sProgram, _Out_ String& sParameters)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

    //Call the low level API
    LPWSTR pwszProgram{nullptr};
    LPWSTR pwszParameters{nullptr};
    const HRESULT hr{p->GetNotifyCmdLine(&pwszProgram, &pwszParameters)};
    if (FAILED(hr))
      return hr;

    //Handle updating the output parameter
    _HandleStringParameter(pwszProgram, sProgram);
    _HandleStringParameter(pwszParameters, sParameters);

    return hr;
  }

  HRESULT GetReplyData(_Out_ ByteArray& buffer)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

    //Call the low level API
    BYTE* pBuffer{nullptr};
    UINT64 nSize{0};
    const HRESULT hr{p->GetReplyData(&pBuffer, &nSize)};
    if (FAILED(hr))
      return hr;

    //Copy the data to the output parameter
    buffer.clear();
#pragma warning(suppress: 26444 26481 26489)
    buffer.insert(buffer.begin(), pBuffer, pBuffer + nSize);
    CoTaskMemFree(pBuffer);

    return hr;
  }

  HRESULT GetReplyFileName(_Out_ String& sReplyFileName)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

    //Call the low level API
    LPWSTR pwszReplyFileName{nullptr};
    const HRESULT hr{p->GetReplyFileName(&pwszReplyFileName)};
    if (FAILED(hr))
      return hr;

    //Handle updating the output parameter
    _HandleStringParameter(pwszReplyFileName, sReplyFileName);

    return hr;
  }
};

//Wrapper for IBackgroundCopyJob3
class CJob3 : public ATL::CComPtr<IBackgroundCopyJob3>
{
//No Helper methods defined
};

//Wrapper for IBackgroundCopyJob4
class CJob4 : public ATL::CComPtr<IBackgroundCopyJob4>
{
//No Helper methods defined
};

//Wrapper for IBackgroundCopyJob5
__if_exists(IBackgroundCopyJob5)
{
class CJob5 : public ATL::CComPtr<IBackgroundCopyJob5>
{
//No Helper methods defined
};
} //__if_exists(IBackgroundCopyJob5)

//Wrapper for IBackgroundCopyJobHttpOptions
class CJobOptions : public ATL::CComPtr<IBackgroundCopyJobHttpOptions>
{
public:
//Methods
#pragma warning(suppress: 26812)
  HRESULT GetClientCertificate(__RPC__out BG_CERT_STORE_LOCATION& StoreLocation, _Out_ String& sStoreName, _Out_ ByteArray& certHashBlob, _Out_ String& sSubjectName)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

    sStoreName.clear();
    sSubjectName.clear();
    certHashBlob.clear();

    //Call the low level API
    LPWSTR pwszStoreName{nullptr};
    BYTE* pCertHashBlob{nullptr};
    LPWSTR pwszSubjectName{nullptr};
    const HRESULT hr{p->GetClientCertificate(&StoreLocation, &pwszStoreName, &pCertHashBlob, &pwszSubjectName)};
    if (FAILED(hr) || (hr == S_FALSE))
      return hr;

    //Copy the data to the output parameters
    certHashBlob.clear();
#pragma warning(suppress: 26444 26481 26489)
    certHashBlob.insert(certHashBlob.begin(), pCertHashBlob, pCertHashBlob + 20);
    _HandleStringParameter2(pwszStoreName, sStoreName);
    _HandleStringParameter2(pwszSubjectName, sSubjectName);

    //Free the memory before we return
    CoTaskMemFree(pCertHashBlob);

    return hr;
  }

  HRESULT GetCustomHeaders(_Out_ String& sRequestHeaders)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

    //Call the low level API
    LPWSTR pwszRequestHeaders{nullptr};
    const HRESULT hr{p->GetCustomHeaders(&pwszRequestHeaders)};
    if (FAILED(hr))
        return hr;

    //Handle updating the output parameter
    _HandleStringParameter(pwszRequestHeaders, sRequestHeaders);

    return hr;
  }
};

//Wrapper for IBackgroundCopyJobHttpOptions2
#if (NTDDI_VERSION >= NTDDI_WIN10_RS5)
__if_exists(IBackgroundCopyJobHttpOptions2)
{
class CJobOptions2 : public ATL::CComPtr<IBackgroundCopyJobHttpOptions2>
{
public:
  HRESULT GetHttpMethod(_Out_ String& sMethod)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

    //Call the low level API
    LPWSTR pwszMethod{nullptr};
    const HRESULT hr{p->GetHttpMethod(&pwszMethod)};
    if (FAILED(hr))
      return hr;

    //Handle updating the output parameter
    _HandleStringParameter(pwszMethod, sMethod);

    return hr;
  }
};
} //__if_exists(IBackgroundCopyJobHttpOptions2)
#endif //#if (NTDDI_VERSION >= NTDDI_WIN10_RS5)

//Wrapper for IBackgroundCopyJobHttpOptions3
__if_exists(IBackgroundCopyJobHttpOptions3)
{
class CJobOptions3 : public ATL::CComPtr<IBackgroundCopyJobHttpOptions3>
{
public:
  template <class ServerCertificateValidationCallback>
  HRESULT SetServerCertificateValidationInterface()
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

#pragma warning(suppress: 26429)
    ATL::CComObject<ServerCertificateValidationCallback>* pSinkEvent{nullptr};
    HRESULT hr{ATL::CComObject<ServerCertificateValidationCallback>::CreateInstance(&pSinkEvent)};
    if (FAILED(hr))
      return hr;
    ATL::CComPtr<IUnknown> unknw;
    hr = pSinkEvent->QueryInterface(&unknw);
    if (FAILED(hr))
    {
      pSinkEvent->Release();
      return hr;
    }
    return p->SetServerCertificateValidationInterface(unknw);
  }
};
} //__if_exists(IBackgroundCopyJobHttpOptions3)

//Wrapper for IBitsTokenOptions
__if_exists(IBitsTokenOptions)
{
class CTokenOptions : public ATL::CComPtr<IBitsTokenOptions>
{
public:
//Methods
  HRESULT GetHelperTokenSid(_Out_ String& sSid)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

    //Call the low level API
    LPWSTR pwszSid{nullptr};
    const HRESULT hr{p->GetHelperTokenSid(&pwszSid)};
    if (FAILED(hr))
      return hr;

    //Handle updating the output parameter
    _HandleStringParameter(pwszSid, sSid);

    return hr;
  }
};
} //__if_exists(IBitsTokenOptions)

//Wrapper for IBackgroundCopyJob
class CJob : public ATL::CComPtr<IBackgroundCopyJob>
{
public:
//Methods
  template <class BackgroundCopyCallback>
  HRESULT SetNotifyInterface()
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

#pragma warning(suppress: 26429)
    ATL::CComObject<BackgroundCopyCallback>* pSinkEvent{nullptr};
    HRESULT hr{ATL::CComObject<BackgroundCopyCallback>::CreateInstance(&pSinkEvent)};
    if (FAILED(hr))
      return hr;
    ATL::CComPtr<IUnknown> unknw;
    hr = pSinkEvent->QueryInterface(&unknw);
    if (FAILED(hr))
    {
      pSinkEvent->Release();
      return hr;
    }
    return p->SetNotifyInterface(unknw);
  }

  HRESULT EnumFiles(_Out_ FilesArray& files)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

    //Get the enumerator
    ATL::CComPtr<IEnumBackgroundCopyFiles> enumFiles;
    HRESULT hr{p->EnumFiles(&enumFiles)};
    if (FAILED(hr))
      return hr;

    //Clear the vector
    files.clear();

    //Work out how many files there are
    ULONG nFiles{0};
    hr = enumFiles->GetCount(&nFiles);
    if (FAILED(hr))
      return hr;

    //Allocate the vector
    files.resize(nFiles);

    //Transfer the COM interfaces
#pragma warning(suppress: 26489)
    for (auto& file : files)
    {
      hr = enumFiles->Next(1, &file, nullptr);
      if (FAILED(hr))
        return hr;
    }

    return hr;
  }

  HRESULT GetJob2(_Inout_ CJob2& job2) noexcept
  {
    return QueryInterface(&job2);
  }

  HRESULT GetJob3(_Inout_ CJob3& job3) noexcept
  {
    return QueryInterface(&job3);
  }

  HRESULT GetJob4(_Inout_ CJob4& job4) noexcept
  {
    return QueryInterface(&job4);
  }

__if_exists(CJob5)
{
  HRESULT GetJob5(_Inout_ CJob5& job5) noexcept
  {
    return QueryInterface(&job5);
  }
} //__if_exists(CJob5)

__if_exists(CTokenOptions)
{
  HRESULT GetTokenOptions(_Inout_ CTokenOptions& options) noexcept
  {
    return QueryInterface(&options);
  }
} //__if_exists(CTokenOptions)

  HRESULT GetJobOptions(_Inout_ CJobOptions& options) noexcept
  {
    return QueryInterface(&options);
  }

__if_exists(CJobOptions2)
{
  HRESULT GetJobOptions2(_Inout_ CJobOptions2& options) noexcept
  {
    return QueryInterface(&options);
  }
} //__if_exists(CJobOptions2)

__if_exists(CJobOptions3)
{
  HRESULT GetJobOptions3(_Inout_ CJobOptions3& options) noexcept
  {
    return QueryInterface(&options);
  }
} //__if_exists(CJobOptions3)

  HRESULT GetDescription(_Out_ String& sDescription)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

    //Call the low level API
    LPWSTR pwszDescription{nullptr};
    const HRESULT hr{p->GetDescription(&pwszDescription)};
    if (FAILED(hr))
      return hr;

    //Handle updating the output parameter
    _HandleStringParameter(pwszDescription, sDescription);

    return hr;
  }

  HRESULT GetDisplayName(_Out_ String& sDisplayName)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

    //Call the low level API
    LPWSTR pwszDisplayName{nullptr};
    const HRESULT hr{p->GetDisplayName(&pwszDisplayName)};
    if (FAILED(hr))
      return hr;

    //Handle updating the output parameter
    _HandleStringParameter(pwszDisplayName, sDisplayName);

    return hr;
  }

  HRESULT GetOwner(_Out_ String& sOwner)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

    //Call the low level API
    LPWSTR pwszOwner{nullptr};
    const HRESULT hr{p->GetOwner(&pwszOwner)};
    if (FAILED(hr))
      return hr;

    //Handle updating the output parameter
    _HandleStringParameter(pwszOwner, sOwner);

    return hr;
  }

#pragma warning(suppress: 26812)
  HRESULT GetProxySettings(_Out_ BG_JOB_PROXY_USAGE& proxyUsage, _Out_ String& sProxyList, _Out_ String& sProxyBypassList)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

    //Call the low level API
    LPWSTR pwszProxyList{nullptr};
    LPWSTR pwszProxyBypassList{nullptr};
    const HRESULT hr{p->GetProxySettings(&proxyUsage, &pwszProxyList, &pwszProxyBypassList)};
    if (FAILED(hr))
      return hr;

    //Handle updating the output parameters
    _HandleStringParameter(pwszProxyList, sProxyList);
    _HandleStringParameter(pwszProxyBypassList, sProxyBypassList);

    return hr;
  }
};

//Wrapper for IBackgroundCopyError
class CError : public ATL::CComPtr<IBackgroundCopyError>
{
public:
//Methods
  HRESULT GetErrorContextDescription(_Out_ String& sContextDescription, _In_ DWORD dwLanguageID = MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT))
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

    //Call the low level API
    LPWSTR pwszContextDescription{nullptr};
    const HRESULT hr{p->GetErrorContextDescription(dwLanguageID, &pwszContextDescription)};
    if (FAILED(hr))
      return hr;

    //Handle updating the output parameter
    _HandleStringParameter(pwszContextDescription, sContextDescription);

    return hr;
  }

  HRESULT GetErrorDescription(_Out_ String& sErrorDescription, _In_ DWORD dwLanguageID = MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT))
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

    //Call the low level API
    LPWSTR pwszErrorDescription{nullptr};
    const HRESULT hr{p->GetErrorDescription(dwLanguageID, &pwszErrorDescription)};
    if (FAILED(hr))
      return hr;

    //Handle updating the output parameter
    _HandleStringParameter(pwszErrorDescription, sErrorDescription);

    return hr;
  }

  HRESULT GetProtocol(_Out_ String& sProtocol)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

    //Call the low level API
    LPWSTR pwszProtocol{nullptr};
    const HRESULT hr{p->GetProtocol(&pwszProtocol)};
    if (FAILED(hr))
      return hr;

    //Handle updating the output parameter
    _HandleStringParameter(pwszProtocol, sProtocol);

    return hr;
  }
};

//The main wrapper for the IBackgroundCopyCallback interface
class ATL_NO_VTABLE CCallback : public ATL::CComObjectRootEx<ATL::CComSingleThreadModel>,
                                public IBackgroundCopyCallback
{
public:
#pragma warning(suppress: 26455)
  CCallback() = default;
  CCallback(const CCallback&) = delete;
  CCallback(CCallback&&) = delete;
  virtual ~CCallback() = default;
  CCallback& operator=(const CCallback&) = delete;
  CCallback& operator=(CCallback&&) = delete;

#pragma warning(suppress: 26440)
  DECLARE_NO_REGISTRY()

#pragma warning(suppress: 26429 26440 26447 26477 26481)
  BEGIN_COM_MAP(CCallback) //NOLINT(modernize-avoid-c-arrays, modernize-use-noexcept)
    COM_INTERFACE_ENTRY(IBackgroundCopyCallback)
#pragma warning(suppress: 26485)
  END_COM_MAP()

//IBackgroundCopyCallback
#pragma warning(suppress: 26433)
  STDMETHOD(JobError)(__RPC__in_opt IBackgroundCopyJob* pJob, __RPC__in_opt IBackgroundCopyError* pError)
  {
    CJob job;
    job.Attach(pJob);
    CError error;
    error.Attach(pError);
    const HRESULT hr{OnJobError(job, error)};
    error.Detach();
    job.Detach();
    return hr;
  }

#pragma warning(suppress: 26433)
  STDMETHOD(JobModification)(__RPC__in_opt IBackgroundCopyJob* pJob, _In_ DWORD dwReserved)
  {
    CJob job;
    job.Attach(pJob);
    const HRESULT hr{OnJobModification(job, dwReserved)};
    job.Detach();
    return hr;
  }

#pragma warning(suppress: 26433)
  STDMETHOD(JobTransferred)(__RPC__in_opt IBackgroundCopyJob* pJob)
  {
    CJob job;
    job.Attach(pJob);
    const HRESULT hr{OnJobTransferred(job)};
    job.Detach();
    return hr;
  }

  //Normalized C++ callback methods
#pragma warning(suppress: 26440)
  virtual HRESULT OnJobError(_In_ CJob& /*job*/, _In_ CError& /*error*/)
  {
    return S_OK;
  }

#pragma warning(suppress: 26440)
  virtual HRESULT OnJobModification(_In_ CJob& /*job*/, _In_ DWORD /*dwReserved*/)
  {
    return S_OK;
  }

#pragma warning(suppress: 26440)
  virtual HRESULT OnJobTransferred(_In_ CJob& /*job*/)
  {
    return S_OK;
  }
};

//Implementation class for IBackgroundCopyCallback2
__if_exists(IBackgroundCopyCallback2)
{
//The main wrapper for the IBackgroundCopyCallback2 interface
class ATL_NO_VTABLE CCallback2 : public ATL::CComObjectRootEx<ATL::CComSingleThreadModel>,
                                 public IBackgroundCopyCallback2
{
public:
  CCallback2() = default;
  CCallback2(const CCallback2&) = delete;
  CCallback2(CCallback2&&) = delete;
  virtual ~CCallback2() = default;
  CCallback2& operator=(const CCallback2&) = delete;
  CCallback2& operator=(CCallback2&&) = delete;

#pragma warning(suppress: 26440)
  DECLARE_NO_REGISTRY()

#pragma warning(suppress: 26429 26440 26447 26477 26481)
  BEGIN_COM_MAP(CCallback2) //NOLINT(modernize-avoid-c-arrays, modernize-use-noexcept)
    COM_INTERFACE_ENTRY(IBackgroundCopyCallback)
    COM_INTERFACE_ENTRY(IBackgroundCopyCallback2)
#pragma warning(suppress: 26485)
  END_COM_MAP()

//IBackgroundCopyCallback
#pragma warning(suppress: 26433)
  STDMETHOD(JobError)(__RPC__in_opt IBackgroundCopyJob* pJob, __RPC__in_opt IBackgroundCopyError* pError)
  {
    CJob job;
    job.Attach(pJob);
    CError error;
    error.Attach(pError);
    const HRESULT hr{OnJobError(job, error)};
    error.Detach();
    job.Detach();
    return hr;
  }

#pragma warning(suppress: 26433)
  STDMETHOD(JobModification)(__RPC__in_opt IBackgroundCopyJob* pJob, _In_ DWORD dwReserved)
  {
    CJob job;
    job.Attach(pJob);
    const HRESULT hr{OnJobModification(job, dwReserved)};
    job.Detach();
    return hr;
  }

#pragma warning(suppress: 26433)
  STDMETHOD(JobTransferred)(__RPC__in_opt IBackgroundCopyJob* pJob)
  {
    CJob job;
    job.Attach(pJob);
    const HRESULT hr{OnJobTransferred(job)};
    job.Detach();
    return hr;
  }

//IBackgroundCopyCallback2
#pragma warning(suppress: 26433)
  STDMETHOD(FileTransferred)(__RPC__in_opt IBackgroundCopyJob* pJob, __RPC__in_opt IBackgroundCopyFile* pFile)
  {
    CJob job;
    job.Attach(pJob);
    CFile file;
    file.Attach(pFile);
    const HRESULT hr{OnFileTransferred(job, file)};
    file.Detach();
    job.Detach();
    return hr;
  }

//Normalized C++ callback methods
#pragma warning(suppress: 26440)
  virtual HRESULT OnJobError(_In_ CJob& /*job*/, _In_ CError& /*error*/)
  {
    return S_OK;
  }

#pragma warning(suppress: 26440)
  virtual HRESULT OnJobModification(_In_ CJob& /*job*/, _In_ DWORD /*dwReserved*/)
  {
    return S_OK;
  }

#pragma warning(suppress: 26440)
  virtual HRESULT OnJobTransferred(_In_ CJob& /*job*/)
  {
    return S_OK;
  }

#pragma warning(suppress: 26440)
  virtual HRESULT OnFileTransferred(_In_ CJob& /*job*/, _In_ CFile& /*file*/)
  {
    return S_OK;
  }
};
} //__if_exists(IBackgroundCopyCallback2)

__if_exists(IBackgroundCopyCallback3)
{
//The main wrapper for the IBackgroundCopyCallback3 interface
class ATL_NO_VTABLE CCallback3 : public ATL::CComObjectRootEx<ATL::CComSingleThreadModel>,
                                 public IBackgroundCopyCallback3
{
public:
  CCallback3() = default;
  CCallback3(const CCallback3&) = delete;
  CCallback3(CCallback3&&) = delete;
  virtual ~CCallback3() = default;
  CCallback3& operator=(const CCallback3&) = delete;
  CCallback3& operator=(CCallback3&&) = delete;

#pragma warning(suppress: 26440)
  DECLARE_NO_REGISTRY()

#pragma warning(suppress: 26429 26440 26447 26477 26481)
  BEGIN_COM_MAP(CCallback3) //NOLINT(modernize-avoid-c-arrays, modernize-use-noexcept)
    COM_INTERFACE_ENTRY(IBackgroundCopyCallback)
    COM_INTERFACE_ENTRY(IBackgroundCopyCallback2)
    COM_INTERFACE_ENTRY(IBackgroundCopyCallback3)
#pragma warning(suppress: 26485)
  END_COM_MAP()

//IBackgroundCopyCallback
#pragma warning(suppress: 26433)
  STDMETHOD(JobError)(__RPC__in_opt IBackgroundCopyJob* pJob, __RPC__in_opt IBackgroundCopyError* pError)
  {
    CJob job;
    job.Attach(pJob);
    CError error;
    error.Attach(pError);
    const HRESULT hr{OnJobError(job, error)};
    error.Detach();
    job.Detach();
    return hr;
  }

#pragma warning(suppress: 26433)
  STDMETHOD(JobModification)(__RPC__in_opt IBackgroundCopyJob* pJob, _In_ DWORD dwReserved)
  {
    CJob job;
    job.Attach(pJob);
    const HRESULT hr{OnJobModification(job, dwReserved)};
    job.Detach();
    return hr;
  }

#pragma warning(suppress: 26433)
  STDMETHOD(JobTransferred)(__RPC__in_opt IBackgroundCopyJob* pJob)
  {
    CJob job;
    job.Attach(pJob);
    const HRESULT hr{OnJobTransferred(job)};
    job.Detach();
    return hr;
  }

//IBackgroundCopyCallback2
#pragma warning(suppress: 26433)
  STDMETHOD(FileTransferred)(__RPC__in_opt IBackgroundCopyJob* pJob, __RPC__in_opt IBackgroundCopyFile* pFile)
  {
    CJob job;
    job.Attach(pJob);
    CFile file;
    file.Attach(pFile);
    const HRESULT hr{OnFileTransferred(job, file)};
    file.Detach();
    job.Detach();
    return hr;
  }

//IBackgroundCopyCallback3
#pragma warning(suppress: 26433 26485)
  STDMETHOD(FileRangesTransferred)(__RPC__in_opt IBackgroundCopyJob* pJob, __RPC__in_opt IBackgroundCopyFile* pFile,
                                   DWORD rangeCount, __RPC__in_ecount_full(rangeCount) const BG_FILE_RANGE ranges[]) //NOLINT(modernize-avoid-c-arrays)
  {
    CJob job;
    job.Attach(pJob);
    CFile file;
    file.Attach(pFile);
    const HRESULT hr{OnFileRangesTransferred(job, file, rangeCount, ranges)};
    file.Detach();
    job.Detach();
    return hr;
  }

//Normalized C++ callback methods
#pragma warning(suppress: 26440)
  virtual HRESULT OnJobError(_In_ CJob& /*job*/, _In_ CError& /*error*/)
  {
    return S_OK;
  }

#pragma warning(suppress: 26440)
  virtual HRESULT OnJobModification(_In_ CJob& /*job*/, _In_ DWORD /*dwReserved*/)
  {
    return S_OK;
  }

#pragma warning(suppress: 26440)
  virtual HRESULT OnJobTransferred(_In_ CJob& /*job*/)
  {
    return S_OK;
  }

#pragma warning(suppress: 26440)
  virtual HRESULT OnFileTransferred(_In_ CJob& /*job*/, _In_ CFile& /*file*/)
  {
    return S_OK;
  }

#pragma warning(suppress: 26440 26485)
  virtual HRESULT OnFileRangesTransferred(_In_ CJob& /*job*/, _In_ CFile& /*file*/, _In_ DWORD rangeCount, __RPC__in_ecount_full(rangeCount) const BG_FILE_RANGE ranges[]) //NOLINT(modernize-avoid-c-arrays)
  {
    UNREFERENCED_PARAMETER(rangeCount);
    UNREFERENCED_PARAMETER(ranges);
    return S_OK;
  }
};
} //__if_exists(IBackgroundCopyCallback3)

__if_exists(IBackgroundCopyServerCertificateValidationCallback)
{
//The main wrapper for the IBackgroundCopyServerCertificateValidationCallback interface
class ATL_NO_VTABLE CServerCertificateValidationCallback : public ATL::CComObjectRootEx<ATL::CComSingleThreadModel>,
                                                           public IBackgroundCopyServerCertificateValidationCallback
{
public:
  CServerCertificateValidationCallback() = default;
  CServerCertificateValidationCallback(const CCallback&) = delete;
  CServerCertificateValidationCallback(CCallback&&) = delete;
  ~CServerCertificateValidationCallback() = default
  CServerCertificateValidationCallback& operator=(const CServerCertificateValidationCallback&) = delete;
  CServerCertificateValidationCallback& operator=(CServerCertificateValidationCallback&&) = delete;

#pragma warning(suppress: 26440)
  DECLARE_NO_REGISTRY()

#pragma warning(suppress: 26429 26440 26477 26481)
  BEGIN_COM_MAP(CServerCertificateValidationCallback)
    COM_INTERFACE_ENTRY(IBackgroundCopyServerCertificateValidationCallback)
#pragma warning(suppress: 26485)
  END_COM_MAP()

//IBackgroundCopyServerCertificateValidationCallback
#pragma warning(suppress: 26433 26485)
  STDMETHOD(ValidateServerCertificate)(__RPC__in_opt IBackgroundCopyJob* job, __RPC__in_opt IBackgroundCopyFile* file, DWORD certLength,
                                       __RPC__in_ecount_full(certLength) const BYTE certData[], DWORD certEncodingType, DWORD certStoreLength,
                                       __RPC__in_ecount_full(certStoreLength) const BYTE certStoreData[])
  {
    CJob jb;
    jb.Attach(job);
    CFile fle;
    fle.Attach(file);
    const HRESULT hr{OnValidateServerCertificate(jb, fle, certLength, certData, certEncodingType, certStoreLength, certStoreData)};
    fle.Detach();
    jb.Detach();
    return hr;
  }

  //Normalized C++ callback methods
#pragma warning(suppress: 26440 26485)
  virtual HRESULT OnValidateServerCertificate(_In_ CJob& /*job*/, _In_ CFile& /*file*/, DWORD /*certLength*/, const BYTE /*certData*/[], DWORD /*certEncodingType*/, DWORD /*certStoreLength*/, const BYTE /*certStoreData*/[])
  {
    return S_OK;
  }
};
} //__if_exists(IBackgroundCopyServerCertificateValidationCallback)

//Wrapper for IBackgroundCopyManager
class CManager : public ATL::CComPtr<IBackgroundCopyManager>
{
public:
//Methods
  HRESULT Initialize() noexcept
  {
    return CoCreateInstance(CLSID_BackgroundCopyManager);
  }

#pragma warning(suppress: 26812)
  HRESULT CreateJob(__RPC__in LPCWSTR pDisplayName, _In_ BG_JOB_TYPE Type, _Out_ GUID& JobID, _Inout_ CJob& job)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

    return p->CreateJob(pDisplayName, Type, &JobID, &job);
  }

  HRESULT EnumJobs(_In_ DWORD dwFlags, _Out_ JobsArray& jobs)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

    //Get the enumerator
    ATL::CComPtr<IEnumBackgroundCopyJobs> enumJobs;
    HRESULT hr{p->EnumJobs(dwFlags, &enumJobs)};
    if (FAILED(hr))
      return hr;

    //Clear the vector
    jobs.clear();

    //Work out how many jobs there are
    ULONG nJobs{0};
    hr = enumJobs->GetCount(&nJobs);
    if (FAILED(hr))
      return hr;

    //Allocate the vector
    jobs.resize(nJobs);

    //Transfer the COM interfaces
#pragma warning(suppress: 26489)
    for (auto& job: jobs)
    {
      hr = enumJobs->Next(1, &job, nullptr);
      if (FAILED(hr))
        return hr;
    }

    return hr;
  }

  HRESULT GetErrorDescription(_In_ HRESULT hResult, _Out_ String& sErrorDescription, _In_ DWORD dwLanguageID = MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT))
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

    //Call the low level API
    LPWSTR pwszErrorDescription{nullptr};
    const HRESULT hr{p->GetErrorDescription(hResult, dwLanguageID, &pwszErrorDescription)};
    if (FAILED(hr))
      return hr;

    //Handle updating the output parameter
    _HandleStringParameter(pwszErrorDescription, sErrorDescription);

    return hr;
  }

  HRESULT GetJob(__RPC__in REFGUID JobID, _Inout_ CJob& Job)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(p != nullptr);

    return p->GetJob(JobID, &Job);
  }
};

}; //namespace BITS

#endif //#ifndef __BITSWRAPPERS_H__